/*
 * node_utils.c
 *
 *    These are utility functions for a node.
 *
 * (C) Copyright IBM Corp. 2000
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <stdarg.h>
#include "node_utils.h"

/* These come from the MQSI V2 ESP 2 README file */
#define DEFAULT_MSG_CAT L"MQSeriesIntegrator2"
#define DEFAULT_MSG_NUM 0x40000A3F

/*
 * NodeHandleFailure
 *    Handle an error in a node's "Evaluate" routine the
 * way the MQSI documentation recommends. Pass it on to 
 * the "failure" terminal if that terminal is connected.
 * Otherwise, raise an exception. The "err" parameter should
 * be the return code from one of the "cni" or "cci" 
 * MQSI library routines.
 */

void NodeHandleFailure(char* file, int line, int err,
					   CciTerminal* failureTerminal, char* msg,
					   CciMessage* dests, CciMessage* excs, CciMessage* outMsg)
{
  if (cniIsTerminalAttached(NULL, failureTerminal))
  {
    NodePrintWarning(file, line,
		             "%s: error = %i", msg, err);
	cniPropagate(NULL, failureTerminal, dests,
		         excs, outMsg);
  }
  else
  {
    if (err == CCI_EXCEPTION)
    {
      cciRethrowLastException(NULL);
    }
    else
	{ 
      NodeThrowFatalException(file, line,
 					        "%s: error = %i", msg, err);
    }
  }
}

/*
 * NodeThrowFatalException
 * NodeThrowConfigException
 *    
 *    These functions throw an exception to MQSI.
 */

void NodeThrowFatalException(char* file,
					int line,
					char* format,
					...)
{
  va_list vParams;
  char msgBuf[1024];

  va_start(vParams, format);
  vsprintf(msgBuf, format, vParams);
  va_end(vParams);
  cciThrowException(NULL, CCI_FATAL_EXCEPTION, file, line,
                    "", DEFAULT_MSG_CAT, DEFAULT_MSG_NUM, msgBuf, NULL);
}

void NodeThrowConfigException(char* file,
					          int line,
					          char* format,
					          ...)
{
  va_list vParams;
  char msgBuf[1024];

  va_start(vParams, format);
  vsprintf(msgBuf, format, vParams);
  va_end(vParams);
  cciThrowException(NULL, CCI_CONFIGURATION_EXCEPTION, file, line,
                    "", DEFAULT_MSG_CAT, DEFAULT_MSG_NUM, msgBuf, NULL);
}

/*
 * NodePrintWarning
 * NodePrintInfo
 *
 *    These functions print a message to the MQSI error
 * log. It's unclear where these messages would actually appear.
 */

void NodePrintWarning(char* file,
			          int line,
				      char* format,
				      ...)
{
  int ignore;
  va_list vParams;
  char msgBuf[1024];

  va_start(vParams, format);
  vsprintf(msgBuf, format, vParams);
  va_end(vParams);
  cciLog(&ignore, CCI_LOG_WARNING, file, line, "", DEFAULT_MSG_CAT,
	     DEFAULT_MSG_NUM, msgBuf, NULL);
}

void NodePrintInfo(char* format,
			   ...)
{
  int ignore;
  va_list vParams;
  char msgBuf[1024];

  va_start(vParams, format);
  vsprintf(msgBuf, format, vParams);
  va_end(vParams);
  cciLog(&ignore, CCI_LOG_INFORMATION, "", 0, "", DEFAULT_MSG_CAT, 
	     DEFAULT_MSG_NUM, msgBuf, NULL);
}

/*
 * NodeGetValue
 *    This routine gets the value of an element given a name.
 * It supports the same syntax as the regular MQSI nodes do for the
 * format of the element name, namely:
 * 
 * Support for "correlation names":
 *    "Root" -- The top of the "root" argument.
 *    "ExceptionList" -- The top of the "exceptionList" argument
 *    "DestinationList" -- The top of the "destinationList" argument
 *    "Body" -- The last child of "Root".
 *
 * Support for arrays, using "[]" syntax.
 * Fields delimeted by a period.
 */

/*
 * nextToken
 *    This routine marks a "NULL" character in the string at the
 * next period, and it also returns the original string. For
 * instance, for the string:
 *    Root.XML.Body.Struct[1].Field
 * It returns:
 *    XML.Body.Struct[1].Field
 * and changes the input string so it points to:
 *    Root
 */

static char* nextToken(char* str, int token)

{
  char* ptr = str;

  if (ptr == NULL)
  {
	return NULL;
  }

  while ((*ptr != '\0') && (*ptr != token))
  {
	ptr++;
  }
  while ((*ptr != '\0') && (*ptr == token))
  {
	*ptr = '\0';
	ptr++;
  }

  if (*ptr == '\0')
  {
	return NULL;
  }
  else
  {
	return ptr;
  }
}

CciElement* getValue(CciElement* elt)

{
  int err;
  CciElement* nextElt;
  CciElementType eltType;

  eltType = cniElementType(&err, elt);
  if ((err == 0) && (eltType & CCI_ELEMENT_TYPE_VALUE))
  {
	/* This is a value or name/value element */
	return elt;
  }
  else if (err == 0)
  {
	/* It's a name element, so take the first child */
    nextElt = cniFirstChild(&err, elt);
	if ((nextElt != NULL) && (err == 0))
	{
	  eltType = cniElementType(&err, nextElt);
	  if ((err == 0) && (eltType & CCI_ELEMENT_TYPE_VALUE))
	  {
		return nextElt;
	  }
	}
  }
  return NULL;
}

#define NAME_BUF_LEN 128

CciElement* findElement(char* name, CciElement* elt, char** nameRemaining)

{
  /*wchar_t nameBuf[NAME_BUF_LEN];UNIX*/
  CciChar nameBuf[NAME_BUF_LEN];
  /*char* aStr = NULL;UNIX*/
  const char* aStr = NULL;
  int err;
  CciElement* curElt;
  CciElement* retVal = NULL;
  char* searchName = strdup(name);
  char* nextName = nextToken(searchName, '.');
  int inc;
  int index;
  char* ptr;

  /*
   * Now, "searchName" points to the name of the first element,
   * and "nextName" points to the rest of the name
   */
  /* set searchName to the name of the first element
   * and ptr to the rest of the array expression */
  ptr = nextToken(searchName, '[');
  if (ptr == NULL)
  {
	/* There's no array subscript */
    index = 0;
  }
  else
  {
	/* There is an array subscript. 
	 * Remember, in MQSI they start with "1" */
	nextToken(ptr, ']');
	index = atoi(ptr) - 1;
  }

  /* Search through children until we find a matching name */
  curElt = cniFirstChild(&err, elt);
  if ((curElt != NULL) && (err == 0))
  {
	cniElementName(&err, curElt, nameBuf, NAME_BUF_LEN);
	if (err == 0)
	{
	  /*aStr = UnicodeToASCII(nameBuf);UNIX*/
	  {
	  const char* mbNameBuf = mbString(nameBuf, BIP_DEF_COMP_CCSID);
	  aStr = mbNameBuf; 
	  }
	}
  }
  while ((err == 0) && (curElt != NULL) && 
	      strcmp(aStr, searchName))
  {
	curElt = cniNextSibling(&err, curElt);
	if (err == 0)
	{
	  cniElementName(&err, curElt, nameBuf, NAME_BUF_LEN);
	  if (err == 0)
	  {
	    free((void*)aStr);
	    /*aStr = UnicodeToASCII(nameBuf);UNIX*/
	    {
	      const char* mbNameBuf = mbString(nameBuf, BIP_DEF_COMP_CCSID);
	      aStr = mbNameBuf; 
	    }
	  }
	}
  }

  /* Continue iterating if there was an array index */
  inc = 0;
  while ((err == 0) && (curElt != NULL) &&
	     (inc < index) && !strcmp(aStr, searchName))
  {
	curElt = cniNextSibling(&err, curElt);
	if (err == 0)
	{
	  cniElementName(&err, curElt, nameBuf, NAME_BUF_LEN);
	  if (err == 0)
	  {
	    free((void*)aStr);
	    /*aStr = UnicodeToASCII(nameBuf);*/
	    {
	      const char* mbNameBuf = mbString(nameBuf, BIP_DEF_COMP_CCSID);
	      aStr = mbNameBuf; 
	    }
	  }
	}
	inc++;
  }
  if (aStr != NULL)
  {
	free((void*)aStr);
  }

  if ((curElt != NULL) && (err == 0) && 
	  ((inc < index) || (index == 0)))
  {
	/* We found an element matching the name */
	if (nextName == NULL)
	{
	  /* We've reached the end of the search */
	  *nameRemaining = NULL;
      retVal = curElt;
	}
	else
	{
	  /* Keep searching */
	  retVal = findElement(nextName, curElt, nameRemaining);
	}
  }
  else
  {
	/* We didn't find a matching name */
	*nameRemaining = strdup(name);
	retVal = elt;
  }

  free(searchName);
  return retVal;
}

CciElement* addChildElement(char* eltName, CciElement* parent)

{
  int err;
  /*wchar_t* uStr;UNIX*/
  const CciChar* uStr;
  CciElement* newElt;
  char* nextName = nextToken(eltName, '.');
  /* 
   * Now, "eltName" points to the name of the element, and
   * "nextName" points to the rest of the name
   */

  /*uStr = ASCIIToUnicode(eltName);UNIX*/
  uStr=CciString(eltName, BIP_DEF_COMP_CCSID) ;
  newElt = cniCreateElementAsLastChild(&err, parent);
  cniSetElementName(&err, newElt, uStr);
  cniSetElementType(&err, newElt, CCI_ELEMENT_TYPE_NAME);
  free((void*)uStr);
  
  if (nextName == NULL)
  {
	/* We've reached the last part of the name,
	 * so we're done */
    return newElt;
  }
  else
  {
	/* Add an intermediate "name" element */
	return addChildElement(nextName, newElt);
  }
}

CciElement* findRootElement(const char* rootName,
						    CciMessage* root, 
						    CciMessage* exceptionList,
						    CciMessage* destinationList)

{
  int err = 0;
  CciElement* firstElt = NULL;

  /* Based on the name, figure out what part of the message
   * to start our search at.
   */
  if (!strcmp(rootName, "Root"))
  {
	firstElt = cniRootElement(&err, root);
  }
  else if (!strcmp(rootName, "ExceptionList"))
  {
	firstElt = cniRootElement(&err, exceptionList);
  }
  else if (!strcmp(rootName, "DestinationList"))
  {
	firstElt = cniRootElement(&err, destinationList);
  }
  else if (!strcmp(rootName, "Body"))
  {
    firstElt = cniRootElement(&err, root);
	if (err == 0)
	{
	  firstElt = cniLastChild(&err, firstElt);
	}
  }

  if (err != 0)
  {
	firstElt = NULL;
  }

  return firstElt;
}

CciElement* NodeGetValue(const char* attrName,
						 CciMessage* root, 
						 CciMessage* exceptionList,
						 CciMessage* destinationList)

{
  char* tmpName = strdup(attrName);
  char* firstBody = nextToken(tmpName, '.');
  CciElement* retVal = NULL;
  CciElement* firstElt;
  CciElement* foundElt;
  char* remaining = NULL;

  /* Find the root element based on the name */
  firstElt = findRootElement(tmpName, root,
						     exceptionList, destinationList);
  if (firstElt != NULL)
  {
	/* Based on the rest of the name, find the element */
    foundElt = findElement(firstBody, firstElt, &remaining);
	if (remaining == NULL)
	{
	  /* We found it, so return the actual element */
	  retVal = getValue(foundElt);
	}
	else
	{
	  free(remaining);
	}
  }

  free(tmpName);
  return retVal;
}

/*
 * NodeAddElement
 *    Add the new element "newElt" to the message at the position
 * specified by "attrName". The naming convention for the name is
 * the same as specified above.
 *    However, we haven't got arrays working yet.
 */

CciElement* NodeAddElement(const char* attrName,
					       CciMessage* root,
					       CciMessage* exceptionList,
					       CciMessage* destinationList)

{
  char* tmpName = strdup(attrName);
  char* firstBody = nextToken(tmpName, '.');
  CciElement* retVal = NULL;
  CciElement* firstElt;
  CciElement* foundElt;
  char* remaining = NULL;

  firstElt = findRootElement(tmpName, root,
						     exceptionList, destinationList);

  if (firstElt != NULL)
  {
    foundElt = findElement(firstBody, firstElt, &remaining);
	if (remaining != NULL)
	{
	  /* It's not there already, so add it */
	  retVal = addChildElement(remaining, foundElt);
	  free(remaining);
	}
  }
  
  free(tmpName);
  return retVal;
}

CciMessage* NodeCopyMessage(CciMessage* msg)

{
  int err;
  int err2;
  CciMessage* retVal;
  CciElement* sourceRoot = NULL;
  CciElement* destRoot = NULL;

  retVal = cniCreateMessage(&err, cniGetMessageContext(&err2, msg));
  if (err == 0)
  {
	sourceRoot = cniRootElement(&err, msg);
  }
  if (err == 0)
  {
    destRoot = cniRootElement(&err, retVal);
  }
  if (err == 0)
  {
	cniCopyElementTree(&err, sourceRoot, destRoot);
  }
  
  if (err == 0)
  {
	return retVal;
  }
  else
  {
	return NULL;
  }
}

/*
 * NodeCreateAttributes
 * NodeDestroyAttributes
 * NodeAddAttribute
 * NodeSetAttribute
 * NodeGetAttribute
 * NodeGetAttributeName
 *
 *    These functions manage an attribute list.
 */

#define INITIAL_ATTR_LIST_SIZE 16

NodeAttributes* NodeCreateAttributes(void)

{
  NodeAttributes* attrs = 
	(NodeAttributes*)malloc(sizeof(NodeAttributes));
  attrs->listSize = INITIAL_ATTR_LIST_SIZE;
  attrs->numAttrs = 0;
  attrs->attrs = 
	(NodeAttribute*)malloc(sizeof(NodeAttribute) * INITIAL_ATTR_LIST_SIZE);
  return attrs;
}

void NodeDestroyAttributes(NodeAttributes* attrs)

{
  unsigned int inc;

  if (attrs != NULL)
  {
	if (attrs->attrs != NULL)
	{
	  for (inc = 0; inc < attrs->numAttrs; inc++)
	  {
		if (attrs->attrs[inc].name != NULL)
		{
		  free(attrs->attrs[inc].name);
		}
		if (attrs->attrs[inc].value != NULL)
		{
		  free(attrs->attrs[inc].value);
		}
	  }
	  free(attrs->attrs);
	}
	free(attrs);
  }
}

/*
 * NodeAddAttribute
 *
 *    This function adds an attribute to the list. Without
 * calling this function first, a call to "NodeSetAttribute"
 * will result in an error.
 */

void NodeAddAttribute(NodeAttributes* attrs, 
					  char* name)
{
  unsigned int attrPos;

  for(attrPos = 0; 
      (attrPos < attrs->numAttrs) && 
	  strcmp(attrs->attrs[attrPos].name, name);
	  attrPos++);
  if (attrPos >= attrs->numAttrs)
  {
	/* Don't set a duplicate */
	if (attrs->numAttrs == attrs->listSize)
	{
	  attrs->listSize *= 2;
	  attrs->attrs = 
		realloc(attrs->attrs, sizeof(NodeAttribute) * attrs->listSize);
	}
    attrs->attrs[attrs->numAttrs].name = strdup(name);
	attrs->attrs[attrs->numAttrs].value = NULL;
	attrs->numAttrs++;
  }
}

/*
 * NodeSetAttribute
 *
 *    If an attribute name is in the list already, then
 * set its value and return 0. Otherwise, return non-zero.
 */

int NodeSetAttribute(NodeAttributes* attrs, 
					 char* name, char* value)
{
  unsigned int attrPos;

  for(attrPos = 0; 
      (attrPos < attrs->numAttrs) && 
	  strcmp(attrs->attrs[attrPos].name, name);
	  attrPos++);
  if (attrPos < attrs->numAttrs)
  {
	/* Attribute already set -- replace the value */
	if (attrs->attrs[attrPos].value != NULL)
	{
	  free(attrs->attrs[attrPos].value);
	}
	attrs->attrs[attrPos].value = strdup(value);
	return 0;
  }
  else
  {
	/* It's not one we've heard of */
	return 99;
  }
}

/*
 * NodeGetAttribute
 *
 *    Given an attribute name, return the attribute, or
 * NULL if the attribute is not in the list or if it
 * is not set.
 */

char* NodeGetAttribute(NodeAttributes* attrs,
						  char* name)
{
  unsigned int attrPos;

  for(attrPos = 0; 
      (attrPos < attrs->numAttrs) && 
	  strcmp(attrs->attrs[attrPos].name, name);
	  attrPos++);
  if (attrPos < attrs->numAttrs)
  {
	/* We found the attribute */
	return attrs->attrs[attrPos].value;
  }
  else
  {
	return NULL;
  }
}

char* NodeGetAttributeName(NodeAttributes* attrs, unsigned int index)

{
  if (index < attrs->numAttrs)
  {
	return (attrs->attrs[index].name);
  }
  else
  {
	return NULL;
  }
}

/*
 * NodeStringBufferAlloc
 * NodeStringBufferFree
 * NodeStringBufferAdd
 * NodeStringBufferAddChar
 *
 *    These functions implement a dynamically-resizable character string.
 * They are used to produce an output message for the parser, since no
 * such mechanism is provided by the broker.
 *    NodeStringBufferAlloc and NodeStringBufferFree allocate memory for the buffer,
 * and free it when necessary.
 *    NodeStringBufferAdd adds a string to the buffer, while NodeStringBufferAddChar
 * adds a single character. These functions automatically resize the
 * buffer when it is necessary, and keep a count of how long the 
 * "relevant" part of the buffer is as well.
 */

#define CHAR_ALLOC_SIZE 32

NodeStringBuffer* NodeStringBufferAlloc(void)

{
  NodeStringBuffer* retVal = (NodeStringBuffer*)malloc(sizeof(NodeStringBuffer));

  retVal->size = CHAR_ALLOC_SIZE;
  retVal->length = 0;
  retVal->buf = (char*)malloc(CHAR_ALLOC_SIZE);
  return retVal;
}

void NodeStringBufferFree(NodeStringBuffer* buf)

{
  free(buf->buf);
  free(buf);
}

static void NodeStringBufferResize(NodeStringBuffer* buf,
								   int len)
{
  if ((buf->length + len + 1) >= buf->size)
  {
	buf->size *= 2;
	if ((buf->length + len + 1) >= buf->size)
	{
	  buf->size += (len + 1);
	}
	buf->buf = (char*)realloc(buf->buf, buf->size);
  }
}

void NodeStringBufferAdd(NodeStringBuffer* buf, const char* str)

{
  int len = strlen(str);

  NodeStringBufferResize(buf, len);
  strcpy(&(buf->buf[buf->length]), str);
  buf->length += len;
}

void NodeStringBufferAddChar(NodeStringBuffer* buf, const char c)

{
  NodeStringBufferResize(buf, 2);
  buf->buf[buf->length] = c;
  buf->length++;
  buf->buf[buf->length] = '\0';
}

/*
 * ASCIIToUnicode
 * UnicodeToASCII
 *    These functions convert between single-byte ASCII strings
 * and double-byte Unicode strings. We will use the standard "mbs"
 * functions if possible, but if these are not available then
 * we will cast the characters instead.
 */

#if HAVE_MBR_FUNCTIONS

wchar_t* ASCIIToUnicode(const char* aStr)

{
  unsigned int len = strlen(aStr);
  wchar_t* outStr = (wchar_t*)malloc(sizeof(wchar_t) * (len + 1));
  mbstowcs(outStr, aStr, len + 1);
  outStr[len] = L'\0';
  return outStr;
}

char* UnicodeToASCII(const wchar_t* lStr)

{
  unsigned int len = wcslen(lStr);
  char* outStr = (char*)malloc(sizeof(char) * (len + 1));
  wcstombs(outStr, lStr, len + 1);
  outStr[len] = '\0';
  return outStr;
}

#else

wchar_t* ASCIIToUnicode(const char* aStr)

{
  wchar_t* retVal;
  char* inPtr;
  wchar_t* outPtr;
  unsigned int len;

  len = strlen(aStr);
  retVal = (wchar_t*)malloc(sizeof(wchar_t) * (len + 1));
  outPtr = retVal;
  inPtr = (char*)aStr;

  while (*inPtr != '\0')
  {
	*outPtr = (wchar_t)*inPtr;
	outPtr++;
    inPtr++;
  }
  *outPtr = L'\0';
  return retVal;
}

char* UnicodeToASCII(const wchar_t* lStr)

{
  char* retVal;
  char* outPtr;
  wchar_t* inPtr;
  unsigned int len;

  len = wcslen(lStr);
  retVal = (char*)malloc(len + 1);
  outPtr = retVal;
  inPtr = (wchar_t*)lStr;
  
  while (*inPtr != L'\0')
  {
	*outPtr = (char)*inPtr;
	outPtr++;
	inPtr++;
  }
  *outPtr = '\0';
  return retVal;
}

#endif /* HAVE_MBR_FUNCTIONS */


CciElement* NodeAddOrGetElement(const char* attrName,
					       CciMessage* root,
					       CciMessage* exceptionList,
					       CciMessage* destinationList)

{
  char* tmpName = strdup(attrName);
  char* firstBody = nextToken(tmpName, '.');
  CciElement* retVal = NULL;
  CciElement* firstElt;
  CciElement* foundElt;
  char* remaining = NULL;

  firstElt = findRootElement(tmpName, root,
						     exceptionList, destinationList);

  if (firstElt != NULL)
  {
    foundElt = findElement(firstBody, firstElt, &remaining);
	retVal=foundElt;
	if (remaining != NULL)
	{
	  /* It's not there already, so add it */
	  /* If script is wrong. and user is trying to insert wrong element*/
	  if(
		  (strcmp(tmpName,"Body")==0 && strcmp(remaining,firstBody)==0) ||
		  (strcmp(tmpName,"Root")==0 && strcmp(remaining,nextToken(firstBody,'.'))==0)
		)
	  {
		retVal=NULL;
	  }
	  else	  
		retVal = addChildElement(remaining, foundElt);
	  free(remaining);
	}
  }
  
  free(tmpName);
  return retVal;
}
