/*
 * node_utils.h
 *
 *    These are utility functions that can be used in MQSI 
 * processing nodes.
 *
 * (C) Copyright IBM Corp. 2000
 */

#ifndef NODE_UTILS_H
#define NODE_UTILS_H

#include <wchar.h>
#include <BipCci.h>
#include <BipCni.h>
#include "DxPoolUtil.h"

#ifdef __cplusplus
extern "C" {
#endif

#define HAVE_MBR_FUNCTIONS 1

/***** Data Structures *****/
 
/*
 * NodeStringBuffer
 *    This can be used to build a dynamic array of characters
 * when a node needs to build up a string bit-by-bit
 */

typedef struct _nodeStringBuffer
{
  int length;
  int size;
  char* buf;
} NodeStringBuffer;

/*
 * NodeAttribute
 * NodeAttributes
 *    These structures hold a set of name-value pairs, and
 * can be used to store configuration attributes for a node.
 */

typedef struct _NodeAttribute

{
  char* name;
  char* value;
} NodeAttribute;

typedef struct _NodeAttributes

{
  NodeAttribute* attrs;
  unsigned int numAttrs;
  unsigned int listSize;
} NodeAttributes;

/***** Functions *****/

/*
 * Exception handling functions
 */

extern void NodeHandleFailure(char* file, int line, int err,
					   CciTerminal* failureTerminal, char* msg,
					   CciMessage* dests, CciMessage* excs, CciMessage* outMsg);
extern void NodeThrowFatalException(char* file,
							        int line,
							        char* format,
							        ...);
extern void NodeThrowConfigException(char* file,
							         int line,
							         char* format,
							         ...);
extern void NodePrintWarning(char* file,
							 int line,
							 char* format,
							 ...);
extern void NodePrintInfo(char* format,
						  ...);

/*
 * Parameter access functions
 */

extern CciElement* NodeGetValue(const char* attrName,
								CciMessage* root, 
								CciMessage* exceptionList,
								CciMessage* destinationList);
extern CciElement* NodeAddElement(const char* attrName,
						          CciMessage* root,
						          CciMessage* exceptionList,
						          CciMessage* destinationList);
extern CciMessage* NodeCopyMessage(CciMessage* msg);

/*
 * String buffer functions
 */

extern NodeStringBuffer* NodeStringBufferAlloc(void);
extern void NodeStringBufferFree(NodeStringBuffer*);
extern void NodeStringBufferAdd(NodeStringBuffer*, const char* str);
extern void NodeStringBufferAddChar(NodeStringBuffer*, const char c);

/*
 * String conversion functions
 */

extern wchar_t* ASCIIToUnicode(const char* aStr);
extern char* UnicodeToASCII(const wchar_t* lStr);

/*
 * Attribute list functions
 */

extern NodeAttributes* NodeCreateAttributes(void);
extern void NodeDestroyAttributes(NodeAttributes*);
extern void NodeAddAttribute(NodeAttributes*, char* name);
extern int NodeSetAttribute(NodeAttributes*, 
							 char* name, char* value);
extern char* NodeGetAttribute(NodeAttributes*,
								 char* name);
extern char* NodeGetAttributeName(NodeAttributes*, unsigned int index);
extern CciElement* NodeAddOrGetElement(const char* attrName,CciMessage* root,CciMessage* exceptionList,CciMessage* destinationList);

#ifdef __cplusplus
}
#endif

#endif
